// Copyright © 2012 Qtrac Ltd.
// 
// This program or package and any associated files are licensed under the
// Apache License, Version 2.0 (the "License"); you may not use these files
// except in compliance with the License. You can get a copy of the License
// at: http://www.apache.org/licenses/LICENSE-2.0.
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package main

import (
    "errors"
    "fmt"
)

type Account struct {
    balance int
}

func (account *Account) Credit(x int) error {
    if x < 0 {
        return errors.New("Negative credit disallowed")
    }
    account.balance += x
    return nil
}

func (account *Account) Debit(x int) error {
    if x < 0 {
        return errors.New("Negative credit disallowed")
    }
    if account.balance-x < 0 {
        return errors.New("Overdraft disallowed")
    }
    account.balance -= x
    return nil
}

func (account *Account) Balance() int {
    return account.balance
}

type UndoableAccount struct {
    Account   // Embedding (delegation)
    functions []func()
}

func (account *UndoableAccount) Credit(x int) error { // Override
    if x < 0 {
        return errors.New("Negative credit disallowed")
    }
    account.balance += x
    account.functions = append(account.functions, func() {
        account.Debit(x)
    })
    return nil
}

func (account *UndoableAccount) Debit(x int) error { // Override
    if x < 0 {
        return errors.New("Negative debit disallowed")
    }
    if account.balance-x < 0 {
        return errors.New("Overdraft disallowed")
    }
    account.balance -= x
    account.functions = append(account.functions, func() {
        account.Credit(x)
    })
    return nil
}

// Exactly the same algorithm as for UndoableIntSet
func (account *UndoableAccount) Undo() error {
    if len(account.functions) == 0 {
        return errors.New("No functions to undo")
    }
    index := len(account.functions) - 1
    account.functions[index]()
    account.functions[index] = nil // Free closure for garbage collection
    account.functions = account.functions[:index]
    return nil
}

func main() {
    account := new(UndoableAccount)
    for _, x := range []int{2, 4, 8, -5} {
        old := account.Balance()
        account.Credit(x)
        fmt.Println(old, "+", x, "=", account.Balance())
    }
    fmt.Println()
    for _, x := range []int{1, -3, 5, 7, 11} {
        old := account.Balance()
        account.Debit(x)
        fmt.Println(old, "-", x, "=", account.Balance())
    }
    fmt.Println()
    for {
        old := account.Balance()
        if err := account.Undo(); err != nil {
            break
        }
        fmt.Println(old, "undo =", account.Balance())
    }
}
