// Copyright © 2012 Qtrac Ltd.
// 
// This program or package and any associated files are licensed under the
// Apache License, Version 2.0 (the "License"); you may not use these files
// except in compliance with the License. You can get a copy of the License
// at: http://www.apache.org/licenses/LICENSE-2.0.
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package main

import (
    "bytes"
    "errors"
    "fmt"
    "sort"
)

type Undo []func()

func (undo *Undo) Add(function func()) {
    *undo = append(*undo, function)
}

func (undo *Undo) Undo() error {
    functions := *undo
    if len(functions) == 0 {
        return errors.New("No functions to undo")
    }
    index := len(functions) - 1
    if function := functions[index]; function != nil {
        function()
        functions[index] = nil // Free closure for garbage collection
    }
    *undo = functions[:index]
    return nil
}

/*
type IntSet struct { // Unwise
    data map[int]bool
    Undo
}
*/
type IntSet struct {
    data map[int]bool
    undo Undo
}

func NewIntSet() IntSet {
    return IntSet{data: make(map[int]bool)}
}

func (set *IntSet) Add(x int) {
    if !set.Contains(x) {
        set.data[x] = true
        set.undo.Add(func() { set.Delete(x) })
    } else {
        set.undo.Add(nil)
    }
}

func (set *IntSet) Delete(x int) {
    if set.Contains(x) {
        delete(set.data, x)
        set.undo.Add(func() { set.Add(x) })
    } else {
        set.undo.Add(nil)
    }
}

func (set *IntSet) Undo() error {
    return set.undo.Undo()
}

func (set *IntSet) Contains(x int) bool {
    return set.data[x]
}

func (set *IntSet) String() string { // Satisfies fmt.Stringer interface
    if len(set.data) == 0 {
        return "{}"
    }
    ints := make([]int, 0, len(set.data))
    for i := range set.data {
        ints = append(ints, i)
    }
    sort.Ints(ints)
    var buffer bytes.Buffer // More efficient than joining lots of little
    buffer.WriteString("{") // strings
    for _, i := range ints {
        fmt.Fprintf(&buffer, "%d,", i)
    }
    return buffer.String()[:buffer.Len()-1] + "}"
}

type Account struct {
    balance int
    undo    Undo
}

func (account *Account) Credit(x int) error {
    if x < 0 {
        return errors.New("Negative credit disallowed")
    }
    account.balance += x
    account.undo.Add(func() { account.Debit(x) })
    return nil
}

func (account *Account) Debit(x int) error {
    if x < 0 {
        return errors.New("Negative credit disallowed")
    }
    if account.balance-x < 0 {
        return errors.New("Overdraft disallowed")
    }
    account.balance -= x
    account.undo.Add(func() { account.Credit(x) })
    return nil
}

func (account *Account) Balance() int {
    return account.balance
}

func (account *Account) Undo() error {
    return account.undo.Undo()
}

func main() {
    ints := NewIntSet()
    for _, i := range []int{1, 3, 5, 7} {
        ints.Add(i)
        fmt.Println(ints)
    }
    for _, i := range []int{1, 2, 3, 4, 5, 6, 7} {
        fmt.Print(i, ints.Contains(i), " ")
        ints.Delete(i)
        fmt.Println(ints)
    }
    for {
        if err := ints.Undo(); err != nil {
            break
        }
        fmt.Println(ints)
    }
    fmt.Println()
    account := new(Account)
    for _, x := range []int{2, 4, 8, -5} {
        old := account.Balance()
        account.Credit(x)
        fmt.Println(old, "+", x, "=", account.Balance())
    }
    fmt.Println()
    for _, x := range []int{1, -3, 5, 7, 11} {
        old := account.Balance()
        account.Debit(x)
        fmt.Println(old, "-", x, "=", account.Balance())
    }
    fmt.Println()
    for {
        old := account.Balance()
        if err := account.Undo(); err != nil {
            break
        }
        fmt.Println(old, "undo =", account.Balance())
    }
    fmt.Println()
    nigel := NewIntSet()
    fmt.Println(nigel)
    nigel.Add(1)
    fmt.Println(nigel)
    nigel.Add(1)
    fmt.Println(nigel)
    nigel.Delete(5)
    fmt.Println(nigel)
    nigel.Delete(1)
    fmt.Println(nigel)
    fmt.Println()
    for {
        if err := nigel.Undo(); err != nil {
            break
        }
        fmt.Println(nigel)
    }
}
